/***
 * mtlib
 *
 * by Marco Tabini
 *
 * This library can be used to track and identify
 * memory leaks.
 *
 * NOTE: This version is thread-safe if a new context
 *       is created for each thread and only used in
 *       that specific thread.
 *
 * mtlib is released into the public domain. Use at your
 * own risk.
***/

#include <stdlib.h>
#include <stdio.h>

#define DEBUG

#include "mtfree.h"

/***
 * insert
 *
 * This function inserts a new MEMDATA entry
 * in the list.
***/

void insert (MT_CONTEXT *context, void *ptr, size_t size, char *filename, long line)
{
	MEMDATA *entry;

	// Create new MEMDATA structure
	// and exit on failure

	entry = malloc (sizeof (MEMDATA));

	if (!entry)
		fprintf (stderr, "Unable to allocate memory in %s:%d", __FILE__, __LINE__);
	else
	{
		// Set the memory block information

		entry->id = context->counter++;
		entry->size = size;
		entry->pointer = ptr;
		entry->next = context->head;

		entry->filename = malloc (strlen (filename) + 1);

		if (entry->filename)
			strcpy (entry->filename, filename);

		entry->line = line;

		// Save the entry in the list

		context->head = entry;
	}

	// If you need to track a particular memory leak,
	// you can set the MT_BREAKPOINT_ID definition to
	// the ID of the memory leak (as printed by
	// mt_terminate()) and set a breakpoint below

#ifdef MT_BREAKPOINT_ID

	if (pointer == MT_BREAKPOINT_ID)
	{
		// set your breakpoint in the next line
		entry = NULL;
	}

#endif

}

/***
 * delete
 *
 * This function deletes an entry from the linked list
***/

void delete (MT_CONTEXT *context, void *ptr)
{
	MEMDATA	*entry, *prev_entry;

	// Only continue if there is a list!

	if (context->head)
	{
		if (context->head->pointer == ptr)
		{
			// If the head of the list is the
			// element that must be deleted,
			// do so now.

			entry = context->head;
			context->head = entry->next;

			if (entry->filename)
				free (entry->filename);
			free (entry);
		}
		else
		{
			// Otherwise, keep looping through
			// the remaining items until the
			// correct one is found.

			prev_entry = context->head;
			entry = context->head->next;

			while (entry)
			{
				if (entry->pointer == ptr)
				{
					// Entry found. Delete it!

					prev_entry->next = entry->next;

					if (entry->filename)
						free (entry->filename);

					free (entry);
					entry = NULL;
				}
				else
				{
					prev_entry = entry;
					entry = entry->next;
				}
			}
		}
	}
}

/***
 * mt_initialize
 *
 * This function creates a new context.
***/

MT_CONTEXT *mt_initialize(char *name)
{
	MT_CONTEXT *context;

	// Create the context and exit the program
	// on error.

	context = malloc (sizeof (MT_CONTEXT));

	if (!context)
	{
		fprintf (stderr, "Unable to start up mtlib in %s:%d.\nAborting...\n\n", __FILE__, __LINE__);
		exit(-1);
	}

	// Set the context information

	context->head = NULL;
	context->counter = 0;
	context->name = malloc (strlen (name) + 1);

	if (context->name)
		strcpy (context->name, name);

	return context;
}

/***
 * mt_terminate
 *
 * This function terminates a context. It should be called at the end
 * of the execution of each thread that creates a context.
 * It will print out a list of all the memory leaks found by mtlib.
***/

void mt_terminate (MT_CONTEXT *context)
{
	MEMDATA *entry, *temp;
	char *filename;

	// Only continue if there are any leaks

	if (context->head)
	{
		// Print out a banner and loop through each leak,
		// printing all the information available.

		fprintf (stderr, "\nWARNING!!!\n");
		fprintf (stderr, "\nPossible memory leaks detected in context %s:\n", ((context->name) ? context->name : "Unknown"));
		fprintf (stderr, "===============================\n");

		entry = context->head;

		while (entry)
		{
			filename = (entry->filename) ? entry->filename : "Unkown file";

			fprintf (stderr, "ID: %d, size %d bytes, allocated in %s:%d\n", entry->id, entry->size, entry->filename, entry->line);

			temp = entry;
			entry = entry->next;

			free (temp->filename);
			free (temp);
		}

		fprintf (stderr, "\n");
	}

	if (context->name)
		free (context->name);

	free (context);
}

/***
 * mt_malloc
 *
 * This function acts as a drop-in replacement for malloc()
***/

void *mt_malloc (MT_CONTEXT *context, size_t size, char *filename, long line)
{
	void *result;

	result = malloc (size);

	if (result)
		insert (context, result, size, filename, line);

	return result;
}

/***
 * mt_calloc
 *
 * This function acts as a drop-in replacement for calloc()
***/

void *mt_calloc (MT_CONTEXT *context, size_t items, size_t size, char *filename, long line)
{
	void *result;

	result = calloc (items, size);

	if (result)
		insert (context, result, size, filename, line);

	return result;
}

/***
 * mt_realloc
 *
 * This function acts as a drop-in replacement for realloc()
***/

void *mt_realloc (MT_CONTEXT *context, void *ptr, size_t size, char *filename, long line)
{
	MEMDATA	*entry;
	void *result;

	entry = NULL;

	if (context->head)
	{
		if (context->head->pointer == ptr)
			entry = context->head;
		else
		{
			entry = context->head->next;

			while ((entry) && (entry->pointer != ptr))
				entry = entry->next;
		}
	}
	
	result = realloc (ptr, size);

	if (result)
	{
		if (entry)
		{
			entry->pointer = result;
			entry->size = size;
		}
		else
			insert (context, result, size, filename, line);
	}

	return result;
}

/***
 * mt_free
 *
 * This function acts as a drop-in replacement for free()
***/

void *mt_free (MT_CONTEXT *context, void *ptr)
{
	delete (context, ptr);

	free (ptr);
}
