/***
 * mtlib
 *
 * by Marco Tabini
 *
 * This library can be used to track and identify
 * memory leaks.
 *
 * NOTE: This version is NOT thread-safe
 *
 * mtlib is released into the public domain. Use at your
 * own risk.
***/

#include <stdlib.h>
#include <stdio.h>

// The MEMDATA structure is used to create
// a linked list through which mtlib tracks
// memory blocks as they are allocated.

typedef struct __TAG_MEMDATA
{
	unsigned long	id;
	size_t		size;
	void		*pointer;

	void		*next;
} MEMDATA;

// The head of the allocation linked list

MEMDATA *head = NULL;

// The counter used to assign a unique ID to
// each memory block

unsigned long counter = 0;

/***
 * insert
 *
 * This function inserts a new MEMDATA entry
 * in the list.
***/

void insert (void *ptr, size_t size)
{
	MEMDATA *entry;

	// Create new MEMDATA structure
	// and exit on failure

	entry = malloc (sizeof (MEMDATA));

	if (!entry)
		fprintf (stderr, 
			 "Unable to allocate memory in %s:%d", 
			 __FILE__, 
			 __LINE__);
	else
	{
		// Set the memory block information

		entry->id = counter++;
		entry->size = size;
		entry->pointer = ptr;
		entry->next = head;

		// Save the entry in the list

		head = entry;
	}

	// If you need to track a particular memory leak,
	// you can set the MT_BREAKPOINT_ID definition to
	// the ID of the memory leak (as printed by
	// mt_terminate()) and set a breakpoint below

#ifdef MT_BREAKPOINT_ID

	if (pointer == MT_BREAK_ID)
	{
		// set your breakpoint in the next line
		entry = NULL;
	}

#endif

}

/***
 * delete
 *
 * This function deletes an entry from the linked list
***/

void delete (void *ptr)
{
	MEMDATA	*entry, *prev_entry;

	// Only continue if there is a list!

	if (head)
	{
		if (head->pointer == ptr)
		{
			// If the head of the list is the
			// element that must be deleted,
			// do so now.

			entry = head;
			head = entry->next;
			free (entry);
		}
		else
		{
			// Otherwise, keep looping through
			// the remaining items until the
			// correct one is found.

			prev_entry = head;
			entry = head->next;

			while (entry)
			{
				if (entry->pointer == ptr)
				{
					// Entry found. Delete it!

					prev_entry->next = entry->next;
					free (entry);
					entry = NULL;
				}
				else
				{
					prev_entry = entry;
					entry = entry->next;
				}
			}
		}
	}
}

/***
 * mt_terminate
 *
 * This function should be called at the end of your program's execution.
 * It will print out a list of all the memory leaks found by mtlib.
***/

void mt_terminate (void)
{
	MEMDATA *entry, *temp;

	// Only continue if there are any leaks

	if (head)
	{
		// Print out a banner and loop through each leak,
		// printing all the information available.

		fprintf (stderr, "Possible memory leaks detected:\n");
		fprintf (stderr, "===============================\n\n");

		entry = head;

		while (entry)
		{
			fprintf (stderr, "ID: %d, size %d bytes\n", entry->id, entry->size);

			temp = entry;
			entry = entry->next;

			free (temp);
		}
	}
}

/***
 * mt_malloc
 *
 * This function acts as a drop-in replacement for malloc()
***/

void *mt_malloc (size_t size)
{
	void *result;

	result = malloc (size);

	if (result)
		insert (result, size);

	return result;
}

/***
 * mt_calloc
 *
 * This function acts as a drop-in replacement for calloc()
***/

void *mt_calloc (size_t items, size_t size)
{
	void *result;

	result = calloc (items, size);

	if (result)
		insert (result, size);

	return result;
}

/***
 * mt_realloc
 *
 * This function acts as a drop-in replacement for realloc()
***/

void *mt_realloc (void *ptr, size_t size)
{
	MEMDATA	*entry;
	void *result;

	entry = NULL;

	// Try to find the original block and
	// change its pointer and size as appropriate

	if (head)
	{
		if (head->pointer == ptr)
			entry = head;
		else
		{
			entry = head->next;

			while ((entry) && (entry->pointer != ptr))
				entry = entry->next;
		}
	}
	
	result = realloc (ptr, size);

	if (result)
	{
		if (entry)
		{
			entry->pointer = result;
			entry->size = size;
		}
		else
			insert (result, size);
	}

	return result;
}

/***
 * mt_free
 *
 * This function acts as a drop-in replacement for free()
***/

void *mt_free (void *ptr)
{
	delete (ptr);

	free (ptr);
}
